#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XShm.h>
#include "x11iface.h"


/***********************************************************************
 *                                                                     *
 *                               BASICS                                *
 *                                                                     *
 ***********************************************************************/

Display *x_display;
Window   x_root_window;
int      x_screen_number;
uint32_t x_root_width, x_root_height;
uint32_t x_red_mask, x_red_shift,
         x_green_mask, x_green_shift,
         x_blue_mask, x_blue_shift;
int      x_depth;
Visual  *x_visual;

/***********************************************************************
  In order to create images etc, we need to find a suitable visual 
  with depth==24 or depth==32.
 ***********************************************************************/

static uint32_t trailing_zero_bits(uint32_t Y)
{
  uint32_t shif;
  for(shif=0;Y%2==0;shif++) Y= Y>>1;
  return shif;
}

static int get_x_visual()
{
  int depth, ndepths, i;
  int *depths;
  XVisualInfo xvinfo;

  depths= XListDepths(x_display, x_screen_number, &ndepths);
  if (!depths) 
  {
    fprintf(stderr,"can't get depth list for display\n");
    return 1;
  }

  depth=0;
  for(i=0;i<ndepths;i++)
    if (depths[i]==32 || depths[i]==24)
    {
      depth= depths[i];
      break;
    }
  XFree(depths);

  if (!depth)
  {
    fprintf(stderr, "display doesn't support 24 or 32 bit depth.\n");
    return 1;
  }

  x_depth= depth;
  x_visual= NULL;
  if (XMatchVisualInfo(x_display, x_screen_number, depth, TrueColor,&xvinfo))
     x_visual= xvinfo.visual;
  else if (XMatchVisualInfo(x_display, x_screen_number,
                                       depth, DirectColor,&xvinfo))
     x_visual= xvinfo.visual;
  else
  {
    fprintf(stderr,"can't find a 24 or 32 bit TrueColor/DirectColor visual\n");
    return 1;
  }
  x_red_mask= x_visual->red_mask;
  x_red_shift= trailing_zero_bits(x_red_mask);
  x_green_mask= x_visual->green_mask;
  x_green_shift= trailing_zero_bits(x_green_mask);
  x_blue_mask= x_visual->blue_mask;
  x_blue_shift= trailing_zero_bits(x_blue_mask);
  return 0;
}


int init_display(char *name)
{
  x_display= XOpenDisplay(name);

  if (!x_display) 
  {
    //FIXME: log this
    fprintf(stderr,"can't open display\n");
    return 1;
  }
  
  x_screen_number= DefaultScreen(x_display);
  x_root_window= DefaultRootWindow(x_display);

  if (get_x_visual())
  {
    XCloseDisplay(x_display);
    x_display= NULL;
    return 1;
  }

  x_root_width= DisplayWidth(x_display, x_screen_number);
  x_root_height= DisplayHeight(x_display, x_screen_number);

  return 0;
}

Window create_window
   (Window parent, int x,int y, int w, int h,
    unsigned int events, int border, int override)
{
  Window win;
  XSetWindowAttributes attribs;
  unsigned int attrib_mask;

  attrib_mask= 0;

  attribs.background_pixmap= None;
    attrib_mask= CWBackPixmap;
  attribs.override_redirect= override; // True;
    attrib_mask|= CWOverrideRedirect;

  win= XCreateWindow(x_display, parent,
                     x, y, w, h,
                     border, x_depth, InputOutput, x_visual, 
                     attrib_mask, &attribs);

  XSelectInput(x_display, win, events);
  XMapWindow(x_display, win);
  if (override)
    XLowerWindow(x_display,win);
  return win;
}


/***********************************************************************
 *                                                                     *
 *                            EVENT SYSTEM                             *
 *                                                                     *
 ***********************************************************************/


typedef struct 
  { Window window; unsigned int event;
    void *data; void (*func)(void *, XEvent *); }
  event_handler_t;

typedef struct
{
   int key,mod;
   int n_mods;
   int *mods;
   void (*press_func)(void *,XEvent *);
   void (*release_func)(void *,XEvent*);
   void *arg;
} grabkey_t;

static int n_grabkey;
static grabkey_t grabkey[50];
static int grab_exmods[10];
static int grab_nexmods;

static int keyboard_grabbed;
static void (*keyb_press_func)(void *,XEvent *);
static void (*keyb_release_func)(void *,XEvent*);
static void *keyb_arg;

static int n_event_handlers, s_event_handlers;
static event_handler_t *event_handlers;


static event_handler_t *find_event(Window window, unsigned int event);

void register_event_handler(Window window, unsigned int event,
                         void *data, void *func, void**old)
{
  event_handler_t *EV,*CP;
  EV= find_event(window, event);
  if (old) *old= NULL;
  if (EV)
  {
    if (old)
    {
      CP= malloc(sizeof(*CP));
      *CP= *EV;
      *old= CP;
    }
    goto done;
  }

  if (n_event_handlers==s_event_handlers)
  {
    if (s_event_handlers==0)
    {
      s_event_handlers= 20;
      event_handlers= malloc(sizeof(event_handlers[0])*s_event_handlers);
    }
    else
    {
      s_event_handlers+= s_event_handlers/2;
      event_handlers= realloc(event_handlers,
                              sizeof(event_handlers[0])*s_event_handlers);
    }
  }
  EV= event_handlers+n_event_handlers;
  n_event_handlers++;

done:
  EV->window= window;
  EV->event= event;
  EV->data= data;
  EV->func= func;
}

void unregister_all_event_handlers(Window window)
{
  int i;
  for(i=0;i<n_event_handlers;i++)
    if (event_handlers[i].window==window)
    {
      event_handlers[i]= event_handlers[n_event_handlers-1];
      n_event_handlers--; i--;
    }
}

void unregister_event_handler(Window window, unsigned int event, void *old)
{
  event_handler_t *EV,*NE;
  EV= old;
  NE= find_event(window, event);
  if (NE && EV)         *NE= *EV; 
  else if (NE && !EV)   *NE= event_handlers[--n_event_handlers];
  else if (!NE && EV)   register_event_handler(window, event,
                                               EV->data, EV->func,NULL);
  if (EV) free(EV);
}


static event_handler_t *find_event(Window window, unsigned int event)
{
  event_handler_t *EV;
  int i;
  for(i=0;i<n_event_handlers;i++)
  {
    EV= event_handlers+i;
    if (EV->window==window && EV->event==event)
       return EV;
  }
  return NULL;
}


void handle_registered_event(XEvent *event)
{
  event_handler_t *EV;

#if 0
  if (event->xany.window==x_root_window)
  {
#if 0
    if (event->type==ButtonPress) syslog(LOG_USER|LOG_NOTICE,"root press\n");
    if (event->type==ButtonRelease) syslog(LOG_USER|LOG_NOTICE,"root release\n");
#endif
    if (event->type==KeyPress || event->type==KeyRelease)
    {
      int i;
      if (keyboard_grabbed)
      {
          void (*f)(void*,XEvent *);
          if (event->type==KeyPress) f= keyb_press_func;
                                else f= keyb_release_func;
          if (f) (*f)(keyb_arg,event);
          return ;
      }
      for(i=0;i<n_grabkey;i++)
        if (grabkey[i].key==event->xkey.keycode)
        {
          void (*f)(void*,XEvent *);
          if (event->type==KeyPress) f= grabkey[i].press_func;
                                else f= grabkey[i].release_func;
          if (f) (*f)(grabkey[i].arg,event);
          return ;
        }
    }
  }
#endif

  EV= find_event(event->xany.window, event->type);
  if (EV && EV->func)
    (*EV->func)(EV->data, event);
}

void call_event_handler(void *H,XEvent *event)
{
  event_handler_t *EV;
  EV= H;
  if (EV && EV->func)
    (*EV->func)(EV->data, event);
}

void free_event_handler(void *H)
{
  if (H) free(H);
}


static int parse_modname(char *mods, int *Rmod)
{
  int i;
  static const struct { char *name; int modmask; } modnames[]=
  { { "shift", ShiftMask },
    { "lock",  LockMask },
    { "capslock",  LockMask },
    { "control", ControlMask },
    { "mod1", Mod1Mask },
    { "mod2", Mod2Mask },
    { "mod3", Mod3Mask },
    { "mod4", Mod4Mask },
    { "mod5", Mod5Mask },
    { "alt",  Mod1Mask },
    { "numlock", Mod2Mask },
    { NULL, 0 } };
  for(i=0;modnames[i].name;i++)
    if (!strcasecmp(mods,modnames[i].name))
     { *Rmod= modnames[i].modmask; return 0; }
  return 1;
}

static int parse_keyspec(char *keyspec, int *Rmod, int *Rkey)
{
  char *plus;
  int mod;
  KeySym sym;
  KeyCode code;

  mod= 0;
  plus= strchr(keyspec,'+');
  if (plus && plus!=keyspec)
  {
    char *mods;
    mods= strdup(keyspec);
    mods[plus-keyspec]=0;
    if (parse_modname(mods, &mod)) { free(mods); return 1; }
    free(mods);
    keyspec= plus+1;
  }

#if 0
  printf("looking up symbol %s\n",keyspec);
#endif

  sym= XStringToKeysym(keyspec);
  if (sym==NoSymbol) return 1;

  code= XKeysymToKeycode(x_display, sym);
  if (!code) return 1;

  if (Rkey) *Rkey= code;
  if (Rmod) *Rmod= mod;
  return 0;
}

static void make_modifier_list(grabkey_t *gk)
{
  int nemo,i,mod;
 
  mod= gk->mod;

  nemo= grab_nexmods;
  for(i=0;i<nemo;i++)
    if (grab_exmods[i]==mod)
    {
      grab_exmods[i]= grab_exmods[nemo-1];
      grab_exmods[nemo-1]= mod;
      nemo--;
      i--;
    }

  gk->n_mods= (1<<nemo);
  gk->mods= malloc(sizeof(int)*gk->n_mods);

  for(i=0;i<(1<<nemo);i++)
  {
    int M,j;
    M= mod;
    for(j=0;j<nemo;j++)
      if (i&(1<<j)) M|=grab_exmods[j];
    gk->mods[i]= M;
  }
}

int grab_key(char *keyspec, int *Rmod, int *Rkey,
             void *press_func, void *release_func, void *arg)
{
  int mod, key;
  int i;
  grabkey_t *gk;

  if (parse_keyspec(keyspec, &mod, &key)) return 1;
  if (n_grabkey==50) return 1;
  for(i=0;i<n_grabkey;i++)
    if (grabkey[i].key==key && grabkey[i].mod==mod)
       return 1;

  gk= grabkey + n_grabkey;
  n_grabkey++;

  gk->key= key;
  gk->mod= mod;
  gk->press_func= press_func;
  gk->release_func= release_func;
  gk->arg= arg;

  make_modifier_list(gk);
  for(i=0;i<gk->n_mods;i++)
  {
    XGrabKey(x_display, gk->key, gk->mods[i], 
             x_root_window, False, GrabModeAsync, GrabModeAsync);
  }

  if (Rmod) *Rmod= mod;
  if (Rkey) *Rkey= key;
  return 0;
}

void grab_extra_modifiers(char *exmod)
{
  char *copy;
  char *str;
  if (!exmod) exmod= "capslock+numlock";

  copy= str= strdup(exmod);
  while(str[0])
  {
    char *plus,*ne;
    plus= strchr(str,'+');
    if (plus) { *plus= 0; ne= plus+1; }
    else { ne= str+strlen(str); }
    if (!parse_modname(str, grab_exmods+grab_nexmods))
      grab_nexmods++;
    str= ne; 
  }
  free(copy);
}

int grab_keyboard(void *press_func, void *release_func, void *arg)
{
  if (keyboard_grabbed) return 1;
  if (!press_func && !release_func) return 1;
  if (XGrabKeyboard(x_display, x_root_window, False, 
        GrabModeAsync, GrabModeAsync, CurrentTime)!=GrabSuccess) return 1;
  keyboard_grabbed= 1;
  keyb_press_func= press_func;
  keyb_release_func= release_func;
  keyb_arg= arg;
  return 0; 
}

void ungrab_keyboard()
{
  if (!keyboard_grabbed) return ;
  XAllowEvents(x_display, AsyncKeyboard,CurrentTime);
  XUngrabKeyboard(x_display, CurrentTime);
  keyboard_grabbed= 0;
}


/***********************************************************************
 *                                                                     *
 *                           IMAGE HANDLING                            *
 *                                                                     *
 ***********************************************************************/



static int can_use_shared_images();
static ximage_t *ximage_create_simple(uint32_t width,uint32_t height);

ximage_t *create_image(uint32_t width, uint32_t height)
{
  ximage_t *image;
  size_t data_size;
  struct shmid_ds dummy_shmid_ds;

  if (!can_use_shared_images()) goto fail_nolib;

  image= calloc(1,sizeof(*image));
  image->width= width;
  image->height= height;
  image->shared= 1;

  // Create the image on the server
  //
  image->x11image= XShmCreateImage(x_display,x_visual, x_depth, ZPixmap,
                                   NULL, &image->shminfo, width, height);
  if (!image->x11image) goto fail_create;

  image->bpl= image->x11image->bytes_per_line;
  data_size= height*image->bpl;

  // Create the shared memory area
  //
  image->shminfo.shmid= shmget(IPC_PRIVATE, data_size, IPC_CREAT | 0777);
  if (image->shminfo.shmid <0)
  {
    fprintf(stderr,"can't shmget shared memory\n");
    goto fail_shmcreate;
  }
  
  image->shminfo.shmaddr= shmat(image->shminfo.shmid,0,0);
   // we remove the ID now, so that when the program exits, the memory
   // will be automatically freed
   //
  shmctl(image->shminfo.shmid, IPC_RMID, &dummy_shmid_ds); 
  if (image->shminfo.shmaddr==(char*) -1) 
  {
    fprintf(stderr,"can't attach shared memory\n");
    goto fail_shmat;
  }

  // bind the area to the image 
  //
  image->data= (uint32_t*) image->shminfo.shmaddr;
  image->x11image->data= (char*) image->shminfo.shmaddr;
  if (!XShmAttach(x_display, &image->shminfo)) 
  {
    fprintf(stderr,"couldn't attach shared memory to the ximage\n");
    goto fail_attach_x;
  }

  return image;

fail_attach_x:
  shmdt(image->shminfo.shmaddr);
fail_shmat:      // We have already removed the ID
fail_shmcreate:
  XDestroyImage(image->x11image);
fail_create:
  free(image);
fail_nolib:
  return ximage_create_simple(width,height);
}



static ximage_t *ximage_create_simple(uint32_t width,uint32_t height)
{
  ximage_t *image;

  image= calloc(1,sizeof(*image));
  image->width= width;
  image->height= height;
  image->data= malloc(sizeof(uint32_t)*width*height);

  image->x11image= XCreateImage(x_display, x_visual, x_depth,
                                ZPixmap, 0, (char*) image->data,
                                width, height, 32, width*sizeof(uint32_t));
      // FIXME: error checking??
  image->bpl= image->x11image->bytes_per_line;
  image->shared= 0;
  return image;
}

/****************************************************************************
 Destroying shared images is a little bit complicated. If we destroy a shared
 image without syncing with the X server, we eventually get some sort of
 XError since the server tries to do some notification regarding the image
 and it's no longer there. So, I destroy the shared memory after we receive
 some events (i.e. output queue is flushed). For this, we need to call
 handle_image_chores in the event loop, after receiving each event.
 ***************************************************************************/

typedef struct image_chore_t
{
  struct image_chore_t *next;
  int kind;
  void *data;
} image_chore_t;

enum { image_chore_shmdt };

static image_chore_t *image_chores;

void destroy_image(ximage_t *image)
{
  image_chore_t *chore;
  if (image->shared) 
  {
    XShmDetach(x_display, &image->shminfo);
    chore= malloc(sizeof(*chore));
    chore->kind= image_chore_shmdt;
    chore->data= image->shminfo.shmaddr;
    chore->next= image_chores;
    image_chores= chore;
  }
  XDestroyImage(image->x11image);
  free(image);
}

void handle_image_chores()
{
  image_chore_t *chore;
  for(chore=image_chores;image_chores;chore=image_chores)
  {
    image_chores= chore->next;
    switch(chore->kind)
    {
      case image_chore_shmdt: shmdt(chore->data); break;
    }
    free(chore);
  }
}

/************************************************************
  Putting an image on the screen is also handled differently
  for shared and non-shared images
 ***********************************************************/

void put_image(Window window, int dest_x, int dest_y,
               ximage_t *image, int src_x, int src_y,
               int width, int height)
{
  if (src_x >= image->width ) return ;
  if (src_y >= image->height ) return ;
  if (src_x<0) { width+= src_x; src_x= 0; }
  if (src_y<0) { height+= src_y; src_y= 0; }
  
  if (width + src_x> image->width)  width= image->width-src_x;
  if (height+ src_y> image->height) height= image->height-src_y;

  if (image->shared)
    XShmPutImage(x_display, window, DefaultGC(x_display,0),
      image->x11image, src_x, src_y, dest_x, dest_y,
      width, height, False);
  else
    XPutImage(x_display, window, DefaultGC(x_display,0),
      image->x11image, src_x, src_y, dest_x, dest_y,
      width, height);
}

/*******************************************************
  Images in X11 can be in shared memory if the server 
  supports the "MIT-SHM" extension.
 ******************************************************/
static int can_use_shared_images()
{
  static int tried;
  static int can;
  if (!tried)
  {
    int ignore, major, minor;
    Bool pixmaps;
    tried= 1;
    can= XQueryExtension(x_display, "MIT-SHM", &ignore, &ignore, &ignore)
         && XShmQueryVersion(x_display, &major, &minor, &pixmaps);
  }
  return can;
}

